# exlap_e.rb  version 1.3: (coding: euc-jp)  2011/06/29
  # ruby library for xmlss(Excel xml SpreadSheet) | csv | html
  # Copyright (C) T. Yoshiizumi, 2010-2011 All rights reserved.
require "nkf"
require "csv"
require "rexml/document"

# gloval variables
$ExlapBaseC = 'e'  # base jcode(nkf option char): e, s, w
$ExlapCsvC = 's' unless defined?($ExlapCsvC) # jcode for csv(nkf option char)
$ExlapTabLen = 4  unless defined?($ExlapTabLen)  # estated tab width
$ExlapFormulaOut = nil  unless $ExlapFormulaOut  # formula output switch
$ExlapHiddenOut = nil  unless $ExlapHiddenOut  # hidden output switch
$ExlapPhoneticTextOut = nil  unless $ExlapPhoneticTextOut
$ExlapLang = :JAPANESE  unless defined?($ExlapLang)
case $ExlapLang
when :JAPANESE
  $defaultFontName = "ͣ Хå"
when :ENGLISH
  $defaultFontName = "Arial"
else
  $defaultFontName = "Arial"
end
unless defined?($ExlapOs)
  if (defined?(ENV['OS']) and ENV['OS'] =~ /^Windows/i) and
     (defined?(ENV['HOME']) and ENV['HOME'] =~ /^[A-Z]\:/i)
    $ExlapOs = :WIN32
  else
    $ExlapOs = :COMMON
  end
end
if RUBY_VERSION < "1.9"
  if $KCODE[0,1].downcase == 'n'
    $KCODE = ($ExlapBaseC == 'w') ? 'u' : $ExlapBaseC
  end
  $ExlapOrgC = $KCODE
end
$ExlapMshttp = nil
$formula = Hash.new  if $formula.class != Hash

# Hash¤ΤΤ褦˻Ȥ褦ˤ뤿λȤ
module ExlHashStruct
  def method_missing(sym, *arg)
    name = sym.to_s
    sz = arg.size
    if sz == 0
      arg = nil
    elsif sz == 1
      arg = arg[0]
    end
    if name =~ /^(.+)=$/
      key = $1
      res = arg
      if arg
        self[key] = arg
      end
    elsif  self.member?(name)
      res = self[name]
    else
      if %w(Alignment Borders Font Interior PhoneticText).include?(name)
        self[name] = {}.structize
        res = self[name]
      else
        res = nil
      end
    end
    return res
  end

  def mrg(other)
    other.each {|key, val|
      if !self[key]
        self[key] = val
      elsif self[key].class == Hash and val.class == Hash
        self[key].structize
        self[key].mrg(val)
      else
        self[key] = val
      end
    }
    return self
  end
end

class Hash
  def structize
    unless @structize_switch
      extend ExlHashStruct
      @structize_switch = true
    end
    return self
  end
end

# 饹°ʤȥ᥽å(Ū˻)
module ExlPrivate
  # constants
  XmlnsHtml = 'http://www.w3.org/TR/REC-html40'
  XmlnsExcel = 'urn:schemas-microsoft-com:office:excel'
  MaxDay = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
  Wareki = {''=>1867, ''=>1911, ''=>1925, 'ʿ'=>1988}
  Month = %w(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec)
  Default_hash = {
    'Alignment'=>{'Vertical'=>"Center"},
    'Font'=>{'Color'=>"#000000", 'FontName'=>$defaultFontName,
        'Size'=>"11"}}
  Outset_worksheet = "<Worksheet ss:Name=\"__sheet_name__\">"
  Table_attr = {'DefaultRowHeight'=>"13.5", 'DefaultColumnWidth'=>"54",
        'x:FullColumns'=>"1", 'x:FullRows'=>"1"}
  Outset_workbook = <<EOS_BOOK
<?xml version=\"1.0\" encoding=\"UTF-8\"?>
<?mso-application progid=\"Excel.Sheet\"?>
<Workbook xmlns:ss=\"urn:schemas-microsoft-com:office:spreadsheet\" \
xmlns:html=\"#{XmlnsHtml}\" xmlns:x=\"#{XmlnsExcel}\" \
xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns=\"urn:\
schemas-microsoft-com:office:spreadsheet\">
<DocumentProperties xmlns=\"urn:schemas-microsoft-com:office:office\">
<Author>__author__</Author>
<LastAuthor>__last_author__</LastAuthor>
<Created>__created__</Created>
<LastSaved>__last_saved__</LastSaved>
<Company>__company__</Company>
</DocumentProperties>
EOS_BOOK
      ##
  Outset_styles = <<EOS_STYLES
<Styles>
<Style ss:Name=\"Normal\" ss:ID=\"Default\"><Alignment ss:Vertical=\
\"Center\"/><Borders/><Font ss:Size=\"11\" x:CharSet=\"128\" ss:Color=\
\"#000000\" ss:FontName=\"#{$defaultFontName}\" x:Family=\"Modern\"/>\
<Interior/><NumberFormat/><Protection/></Style>
EOS_STYLES
      ##
  Outset_html = <<EOS_HTML
<html>
<head>
<meta http-equiv=\"content-type\" content=\"text/html; charset=__jcode__\">
<title>__title__</title>
</head>
<body>
EOS_HTML


  # ʰnkfƤӽФ
  def xnkf(str, code=nil)
    code = $ExlapBaseC unless code
    return NKF.nkf("-m0#{code}", str)
  end

  # $KCODEΰŪѥ᥽å
  def change_kcode(code = nil)
    if RUBY_VERSION < "1.9"
      if code
        $KCODE = code
      else
        $KCODE = $ExlapOrgC
      end
    end
  end

  # ʸʸΥ
  def str_len(str, opt=:CHAR)
    s = str.dup
    if opt == :CHAR  # ʸñ(ХȤǤϤʤ)
      s.gsub!(/./, "x")
    else  # 
      s.gsub!(/[^\x00-\x7f]/, "xx")
    end
    return s.length
  end

  # Hashοʣ֤
  def hash_copy(obj)
    if (cl = obj.class) != Hash
      return (cl == String) ? obj.dup : obj
    end
    res = Hash.new
    obj.each {|key, val|
      res[key] = hash_copy(val)
    }
    return res
  end

  # ȥ륳ɤüʸ(<, >, &)ʤɤĴ
  def arrange_char(s)
    if RUBY_VERSION < "1.9"
      org_code = $KCODE
      $KCODE = ($ExlapBaseC == 'w') ? 'u' : $ExlapBaseC
    end
    s = xnkf(s)
    s = s.gsub(/\r\n/, "\n")
    s.gsub!(/\n/, "\r\n")
    s.gsub!(/</, "&lt;")
    s.gsub!(/>/, "&gt;")
    s.gsub!(/\&/, "&amp;")
    s.gsub!(/"/, "&quot;")
    if $ExlapLang == :JAPANESE
      s.gsub!(//, "&#65374;")
      s.gsub!(//, "&#65293;")
      s.gsub!(//, "&#8741;")
      s.gsub!(//, "&#65504;")
      s.gsub!(//, "&#65505;")
      s.gsub!(//, "&#65506;")
    end
    if RUBY_VERSION < "1.9"
      s.gsub!(/[\x00-\x1f]/) {|mp|
        mp = "&##{mp[0]};"
      }
    else  # if RUBY_VERSION >= "1.9"
      s.gsub!(/[\x00-\x1f]/) {|mp|
        mp = "&##{mp.getbyte(0)};"
      }
    end
    if RUBY_VERSION < "1.9"
      $KCODE = org_code
    end
    return s
  end

  # htmlüʸ̾
  def recover_char(str, code=$ExlapBaseC)
    unless str =~ /\&\#.+;|\&[a-z]+;/i
      return str
    end
    str = str.gsub(/\&\#x([0-9a-f]+);/i) {|ch|
      n = $1.hex
      ch = "&##{n};"
    }
    str.gsub!(/\&\#(\d+);/) {|ptn|
      n = $1.to_i
      if n > 0xff
        ch = NKF.nkf("-m0W16#{code}", [n].pack("n"))
      else
        ch = n.chr
      end
      ptn = ch
    }
    str.gsub!(/\r\n/, "\n")
    if str =~ /\&[a-z]+;/i
      str.gsub!(/\&nbsp;/i, " ")
      str.gsub!(/\&quot;/i, "\"")
      str.gsub!(/\&lt;/i, "<")
      str.gsub!(/\&gt;/i, ">")
      str.gsub!(/\&amp;/i, "&")
    end
    return str
  end

  # ȤΥץxmlѴ
  def tag_attr(tagname, attr_hs, inner_str=nil)
    xmlsrc = ''
    if attr_hs.class != Hash or attr_hs.size < 1
      if inner_str
        xmlsrc = "<#{tagname}>" + inner_str + "</#{tagname}>"
      else
        xmlsrc = "<#{tagname}/>"
      end
      return xmlsrc
    end
    nest = 0
    attr_hs.each {|key, val|
      if val.class == Hash
        nest += 1
      end
    }
    if nest > 0 and attr_hs.size == nest
      xmlsrc += "<#{tagname}>"
      attr_hs.each {|key, val|
        child_tagname = (tagname =~ /^borders$/i) ? 'Border' : key
        xmlsrc += tag_attr(child_tagname, val)
      }
      xmlsrc += inner_str  if inner_str
      xmlsrc += "</#{tagname}>"
    elsif nest == 0
      xmlsrc += "<#{tagname}"
      attr_hs.each {|key, val|
        key = key.to_s
        key = 'ss:' + key  if key !~ /^[a-z]+\:|^xmlns$/i
        opt = " #{key}=\"#{val}\""
        xmlsrc += opt
      }
      if inner_str
        xmlsrc += ">" + inner_str + "</#{tagname}>"
      else
        xmlsrc += "/>"
      end
    else
      STDERR.puts "Error of the array:\n" +
          "key: #{key}\n" +
          "value:\n" +
          "#{val.to_a}"
    end
    return xmlsrc
  end

  def hash_to_xml(hs)
    xmlsrc = ''
    hs.each {|key, val|
      xmlsrc += tag_attr(key, val)
    }
    return xmlsrc
  end

  # ʸhtml sourceݤȽ
  def is_html(str)
    return (str =~ /^\s*</ and str =~ /<html/i and str =~ /<\/html>/i) ?
        true : false
  end

  # Ϳ줿ʸxmlssݤȽ
  def is_xmlss(str)
    res = false
    if str =~ /\A\s*<\?xml/i and
       str =~ /<Workbook [^>]*xmlns=[^>]+office\:spreadsheet/i and
       str =~ /office\:excel/i and str =~ /<Worksheet[^a-z]/i and
       str =~ /<Table[^a-z]/i and str =~ /<Row[^a-z]/i
        res = true
      end
      return res
  end


  # ʸɤĴ٤
  def array_code(aa)
    jp_charset = [NKF::EUC, NKF::JIS, NKF::SJIS, NKF::UTF8]
    code = nil
    aa.each do |row|
      next  if !row or row.size < 1
      row.each do |col|
        next  unless col
        if col.class == Hash
          if (data = col['Data']) and data.class == String
            wcode = NKF::guess(data)
            if jp_charset.include?(wcode)
              code = wcode
              break
            end
          end
          if (pht = col['PhoneticText'])
            pht = pht['Text']  if pht.class == Hash and pht['Text']
            wcode = NKF::guess(pht.to_s)
            if jp_charset.include?(wcode)
              code = wcode
              break
            end
          end
          if col['Font'] and col['Font']['FontName']
            wcode = NKF::guess(col['Font']['FontName'])
            if jp_charset.include?(wcode)
              code = wcode
              break
            end
          end
        else
          if col and col.class == String
            wcode = NKF::guess(col)
            if jp_charset.include?(wcode)
              code = wcode
              break
            end
          end
        end
      end
      break  if code
    end
    res = nil
    if code
      case code
      when NKF::EUC,
        res = 'e'
      when NKF::JIS
        res = 'j'
      when NKF::SJIS
        res = 's'
      when NKF::UTF8
        res = 'w'
      end
    end
    return res
  end

  def wareki(yy)
    if yy > 1988
      gg = 'ʿ'
      gy = yy - 1988
    elsif yy > 1925
      gg = ''
      gy = yy - 1925
    elsif yy > 1911
      gg = ''
      gy = yy - 1911
    else
      gg = ''
      gy = yy - 1867
    end
    return [gg, gy]
  end

  # NumberFormat˽äơ٤νĴԤ
  def n_format(hs, code)
    data = hs['Data'].to_s.dup
    if $ExlapFormulaOut and (fml = hs['Formula']) and !$formula[fml]
      return "#{data} |#{hs['Formula']}"
    end
    fmt = hs['NumberFormat']
    typ = hs['Type'].to_s.downcase
    pht = hs['PhoneticText']
    pht = pht['Text']  if pht.class == Hash and pht['Text']
    if typ =~ /^(.+)\@[a-z]+$/
      typ = $1
      data = data.gsub(/<[^>]+>/, "")
    end
    if !fmt or ["number", "datetime"].include?(typ) != true
      if $ExlapPhoneticTextOut and pht
        data = "#{data}[#{pht}]"
      end
      return data
    end
    if typ == "number"
      percent = nil
      if fmt =~ /\%$/
        percent = true
        fmt.sub(/\%$/, "")
      end
      pfmt = nil
      if fmt =~ /^0[0#,]*$/
        l = fmt.gsub(/,/, "").length
        pfmt = (l>1) ? sprintf("%%0%dd", l) : "%d"
      elsif fmt =~ /^\#[0#,]*$/
        pfmt = "%d"
      elsif fmt =~ /^[#0][#0,]*\.(0+)$/
        l = $1.length
        pfmt = sprintf("%%.%df", l)
      end
      float_sw = (fmt =~ /\./) ? true : false
      data = percent ? data.to_f * 100.0 : data.to_f
      if data >= 0.0
        data = pfmt ? sprintf(pfmt, data) : data.to_s
      else
        data *= -1.0
        data = pfmt ? sprintf(pfmt, data) : data.to_s
        data = '-' + data
      end
      data.sub!(/\.\d+$/, "")  unless float_sw
      data += '%'  if percent
      if fmt =~ /[#0],[#0]/ and data =~ /\d{4}/  # ޶ڤղ
      if data =~ /^(-?)(\d+)(\.\d+)*(\%?)$/
        mm = $1.to_s
        ii = $2
        ff = $3.to_s
        ww = $4.to_s
        rr = ii.reverse
        while rr =~ /\d{4}/
          rr = rr.sub(/(\d{3})(\d)/, "\\1,\\2")
        end
        ii = rr.reverse
        data = mm + ii + ff + ww
      end
    end
    # ʹߤ DateTime ξ
    elsif data =~ /^(\d{4})-(\d\d?)-(\d\d?)T(\d\d?):(\d\d?):(\d\d?)\.(\d+)$/
      y=$1.to_i;  m=$2.to_i;  d=$3.to_i
      hh=$4.to_i;  mm=$5.to_i;  ss=$6.to_i
      mss = $7.to_i
      if fmt == "Short Time"
        data = sprintf("%02d:%02d", hh, mm)
      elsif fmt == "h:mm:ss"
        data = sprintf("%02d:%02d:%02d", hh, mm, ss)
      elsif fmt == "Short Date"
        case $ExlapLang
        when :JAPANESE
          data = sprintf("%d/%02d/%02d", y, m, d)
        when :ENGLISH
          data = sprintf("%d02/%02d/%d", m, d, y)
        end
      elsif fmt == "Long Date"
        case $ExlapLang
        when :JAPANESE
          data = xnkf(sprintf("%dǯ%d%d", y, m, d), code)
        when :ENGLISH
          data = sprintf("%s %d, %d", Month[m-1], d, y)
        end
      elsif $ExlapLang == :JAPANESE and fmt =~ /^\[JPN\]\[\$-411\]g+y/
        gg, gy = wareki(y)
        data = xnkf(sprintf("%s%dǯ%d%d", gg, gy, m, d), code)
      end
    end
    return data
  end

  # Formulaؽܤ
  def formula_replace(aa)
    fr_code = array_code(aa)
    if RUBY_VERSION < "1.9" and fr_code
      org_code = $KCODE
      $KCODE = (fr_code == 'w') ? 'u' : fr_code
    end
    wformula = Hash.new
    $formula.each {|key, val|
      wformula[xnkf(key, fr_code)] = xnkf(val, fr_code)
    }
    for rn in 0...aa.size
      next  unless aa[rn]
      for cn in 0...aa[rn].size
        next  unless aa[rn][cn]
        next  unless aa[rn][cn].class == Hash
        if (fr_formula = aa[rn][cn]['Formula']) and
           (fr_script = wformula[fr_formula]) and fr_script != ''
          data = nil
          eval(fr_script)
          aa[rn][cn]['Data'] = data  if data
        end
      end
    end
    if RUBY_VERSION < "1.9" and fr_code
      $KCODE = org_code
    end
  end

  # ̾Ĵ
  def arrange_sheet_name(org_name, snames, count)
    res = ''
    if !org_name or org_name == '' or
       org_name =~ /^Sheet\d+$/i
      res = sprintf("Sheet%d", count+1)
    elsif !snames.index(org_name)
      res = org_name
      snames.push(res)
    elsif snames.index(org_name) and str_len(org_name) < 27
      w = ''
      for i in 2..99
        w = org_name + sprintf("(%d)", i)
        break  unless snames.index(w)
      end
      unless snames.index(w)
        res = w
        snames.push(res)
      end
    end
    if res == ''
      res = sprintf("Sheet%d", count+1)
    end
    return res
  end

  # γǤĴ
    # ʸɤbase_code˹碌
    # 裳 :TO_STRING ʤHashString:TO_HASHʤStringHash
  def arrange_array_s(obj, base_code=nil, form=nil)
    ary = (obj.class == ExlSheet) ? obj.array : obj
    aa = Array.new
    ary.size.times do
      aa << Array.new
    end
    for rn in 0...ary.size
      for cn in 0...ary[rn].size
        if (w = ary[rn][cn])
          aa[rn][cn] = hash_copy(w)
        end
      end
    end
    base_code = $ExlapBaseC  unless base_code
    if !form
      code = array_code(aa)
      if !code or code == base_code
        return aa
      end
    end
    if form == :TO_STRING
      if $formula.size > 0
        formula_replace(aa)
      end
      if obj.class == ExlSheet and $ExlapHiddenOut != true
        if obj.row.size > 0
          ra = Array.new(obj.row.size)
          for i in 0...obj.row.size
            if obj.row[i] and (w = obj.row[i]['Hidden']) and w.to_i == 1
              ra[i] = 1
              if (span = obj.row[i]['Span'])
                span = span.to_i
                for j in 1..span
                  ra[i+j] = 1
                end
              end
            end
          end
          for i in (0...ra.size).to_a.reverse
            next  unless ra[i] == 1
            if !$ExlapHiddenOut
              aa[i] = []
            elsif $ExlapHiddenOut == :DELETE
              aa.delete_at(i)
            end
          end
        end
        if obj.col.size > 0
          ca = Array.new(obj.col.size)
          for i in 0...obj.col.size
            if obj.col[i] and (w = obj.col[i]['Hidden']) and w.to_i == 1
              ca[i] = 1
              if (span = obj.col[i]['Span'])
                span = span.to_i
                for j in 1..span
                  ca[i+j] = 1
                end
              end
            end
          end
          for i in (0...ca.size).to_a.reverse
            next  unless ca[i] == 1
            for r in 0...aa.size
              next  if !aa[r] or aa[r].size < i
              if !$ExlapHiddenOut
                aa[r][i] = nil
              elsif $ExlapHiddenOut == :DELETE
                aa[r].delete_at(i)
              end
            end
          end
        end
      end
    end
    for rn in 0...aa.size
      next  unless aa[rn]
      for cn in 0...aa[rn].size
        next  unless aa[rn][cn]
        if aa[rn][cn].class == Hash
          aa[rn][cn].each {|k, v|
            case k
            when 'Data', 'NumberFormat', 'Formula', 'NamedCell'
              aa[rn][cn][k] = xnkf(v.to_s, base_code)  if v
            when 'Font'
              if (w = v['FontName'])
                aa[rn][cn][k]['FontName'] = xnkf(w, base_code)
              end
            when 'PhoneticText'
              if v.class == Hash
                v['Text'] = xnkf(v['Text'].to_s, base_code)  if v['Text']
              else
                v = xnkf(v.to_s, base_code)
              end
              aa[rn][cn][k] = v
            end
          }
          if form == :TO_STRING
            aa[rn][cn] = n_format(aa[rn][cn], base_code)
          end
        else
          if aa[rn][cn]
            str = xnkf(aa[rn][cn].to_s, base_code)
            aa[rn][cn] = (form == :TO_HASH) ? {'Data'=>str} : str
          end
        end
      end
    end
    return aa
  end

  # csvѤѴ
  def arrange_for_csv(*ary)
    csv_ary = Array.new
    ary.each {|aa|
      if aa.class == Array
        csv_ary << arrange_array_s(aa, $ExlapCsvC, :TO_STRING)
      elsif aa.class == ExlSheet and aa.array
        csv_ary << arrange_array_s(aa, $ExlapCsvC, :TO_STRING)
      end
    }
    return csv_ary
  end

  # ץȤʸɤѴ(Array, Hashξˤб)
  def code_adj(obj)
    res = nil
    cl = obj.class
    if cl == Array
      res = []
      obj.each {|elem|
        res << code_adj(elem)
      }
    elsif cl == Hash
      res = {}
      obj.each {|key, val|
        res[key] = code_adj(val)
      }
    else
      res = (cl == String) ? xnkf(obj) : obj
    end
    return res
  end

  def int_t(obj)
    return [Fixnum,Bignum].include?(obj.class)
  end

  def append_default_hash(hs)
    dk = []
    hs.each {|key, val|
      dk.push(key)  if val == {}
    }
    dk.each {|key|  hs.delete(key)}
    Default_hash.each {|key, val|
      next  if hs[key].class != Hash
      val.each {|k, v|
        unless hs[key][k]
          hs[key][k] = v
        end
      }
    }
    if hs['Interior'].class == Hash and hs['Interior']['Color'] and
       !hs['Interior']['Pattern']
      hs['Interior']['Pattern'] = "Solid"
    end
    if hs['Interior'] and !hs['Font']
      hs['Font'] = Default_hash['Font']
    end
  end
end

# 饹°ʤ᥽å()
module Exl
  def arrange_array(obj, base_code=nil, form=nil)
    include ExlPrivate
    return arrange_array_s(obj, base_code, form)
  end

  def convert_for_csv(*ary)
    include ExlPrivate
    one_flag = (ary.size == 1) ? true : false
    csv_ary = arrange_for_csv(*ary)
    return one_flag ? csv_ary[0] : csv_ary
  end

  # Hash֥Ȥ򡢲̳ؤޤƹ¤β
  def hash_structize(obj)
    if obj.class != Hash
      return obj
    end
    obj.structize
    obj.each {|key, val|
      next  unless val.class == Hash
      obj[key] = hash_structize(val)
    }
    return obj
  end

  def abs_rc(y,x)
    return sprintf("R%dC%d", y+1, x+1)
  end

  def rel_rc(y,x)
    return sprintf("R[%d]C[%d]", y,x).gsub(/\[0\]/, "")
  end

  # ʰ׵ѥ᥽å
  def eWidth(n)
    hs = {}
    if n.to_f > 0.0
      hs['Width'] = "#{6.5*n.to_f}"
    end
    return hs.structize
  end

  def eHeight(n)
    hs = {}
    if n.to_f > 0.0
      hs['Height'] = "#{13.5*n.to_f}"
    end
    return hs.structize
  end

  def eBorder(line_style=nil, weight=nil, color=nil)
    line_style = "Continuous"  unless line_style
    weight = 1  unless weight
    res = {'Borders'=>{
      'Border1'=>{"Position"=>"Top",
          "LineStyle"=>"#{line_style}",
          "Weight"=>"#{weight}"},
      'Border2'=>{"Position"=>"Bottom",
          "LineStyle"=>"#{line_style}",
          "Weight"=>"#{weight}"},
      'Border3'=>{"Position"=>"Left",
          "LineStyle"=>"#{line_style}",
          "Weight"=>"#{weight}"},
      'Border4'=>{"Position"=>"Right",
          "LineStyle"=>"#{line_style}",
          "Weight"=>"#{weight}"}}}
    if color
      for n in 1..4
        res["Borders"]["Border#{n}"]["Color"] = color
      end
    end
    return hash_structize(res)
  end
  alias :eBox :eBorder

  def eAlignment(horizontal=nil, vertical=nil, wrap=nil)
    hs = {'Vertical'=>"Center"}
    hs['Horizontal'] = horizontal  if horizontal
    hs['Vertical'] = vertical  if vertical
    hs['WrapText'] = wrap.to_s  if wrap
    return hash_structize({'Alignment'=>hs})
  end

  def eFont(color=nil, font_name=nil, size=nil)
    hs = {'Color'=>"#000000", 'FontName'=>$defaultFontName,
        'Size'=>"11"}
    hs['Color'] = color  if color
    hs['FontName'] = font_name  if font_name
    hs['Size'] = size  if size
    return hash_structize({'Font'=>hs})
   end

  def eInterior(color=nil, pattern=nil)
    hs = {}
    hs['Color'] = color  if color
    hs['Pattern'] = pattern  if pattern
    if hs['Color'] and !hs['Pattern']
      hs['Pattern'] = "Solid"
    end
    return hash_structize((hs.size > 0) ? {'Interior'=>hs} : hs)
  end

  def eSpan(cspan=nil, rspan=nil)
    hs = {}
    hs['Cspan'] = cspan.to_s  if cspan and cspan.to_i > 1
    hs['Rspan'] = rspan.to_s  if rspan and rspan.to_i > 1
    return hs.structize
  end
  alias :eMerge :eSpan

  def eNumberFormat(val=nil)
    hs = {}
    val = val.to_s
    hs['NumberFormat'] = val  if val != ''
    return hs.structize
  end

  def eFormula(val=nil)
    hs = {}
    val = val.to_s
    hs['Formula'] = val  if val != ''
    return hs.structize
  end

  def eType(val=nil)
    hs = {}
    val = val.to_s
    hs['Type'] = val  if val != ''
    return hs.structize
  end

  def eData(obj, *params)
    if !obj
      return nil
    end
    if obj.class != Hash
      obj = {'Data'=>obj.to_s}
    end
    while prm = params.shift
      if prm.class == Hash and prm.size > 0
        obj.update(prm)
      end
    end
    return hash_structize(obj)
  end

  def eHidden(n)
    hs = {}
    if [0,1].include?(n.to_i)
      hs['Hidden'] = n
    end
    return hs.structize
  end

  def ePhoneticText(text=nil, visible=nil)
    hs = {}
    text = text.to_s
    if text != ''
      if visible
        hs['PhoneticText'] = {'Text'=>text, 'Visible'=>visible}
      else
        hs['PhoneticText'] = text
      end
    end
    return hash_structize(hs)
  end

  def eHRef(val=nil)
    hs = {}
    val = val.to_s
    hs['HRef'] = val  if val != ''
    return hs.structize
  end

  def eNamedCell(val=nil)
    hs = {}
    val = val.to_s
    hs['NamedCell'] = val  if val != ''
    return hs.structize
  end

  def eBorder_t(line_style=nil, weight=nil, color=nil)
    line_style = "Continuous"  unless line_style
    weight = 1  unless weight
    res = {'Borders'=>{
      'Border1'=>{"Position"=>"Top",
          "LineStyle"=>"#{line_style}",
          "Weight"=>"#{weight}"}}}
    if color
      res["Borders"]["Border1"]["Color"] = color
    end
    return hash_structize(res)
  end

  def eBorder_b(line_style=nil, weight=nil, color=nil)
    line_style = "Continuous"  unless line_style
    weight = 1  unless weight
    res = {'Borders'=>{
      'Border2'=>{"Position"=>"Bottom",
          "LineStyle"=>"#{line_style}",
          "Weight"=>"#{weight}"}}}
    if color
      res["Borders"]["Border2"]["Color"] = color
    end
    return hash_structize(res)
  end

  def eBorder_l(line_style=nil, weight=nil, color=nil)
    line_style = "Continuous"  unless line_style
    weight = 1  unless weight
    res = {'Borders'=>{
      'Border3'=>{"Position"=>"Left",
          "LineStyle"=>"#{line_style}",
          "Weight"=>"#{weight}"}}}
    if color
      res["Borders"]["Border3"]["Color"] = color
    end
    return hash_structize(res)
  end

  def eBorder_r(line_style=nil, weight=nil, color=nil)
    line_style = "Continuous"  unless line_style
    weight = 1  unless weight
    res = {'Borders'=>{
      'Border4'=>{"Position"=>"Right",
          "LineStyle"=>"#{line_style}",
          "Weight"=>"#{weight}"}}}
    if color
      res["Borders"]["Border4"]["Color"] = color
    end
    return hash_structize(res)
  end

  def getData(obj)
    return (obj.class == Hash) ? obj['Data'] : obj
  end

  def max_rc(ary)
    mr = mc = 0
    return [mr,mc]  if !ary or ary == [] or ary == [[]]
    mr = ary.size
    ary.each {|row|
      next  if !row
      w = row.size
      mc = w  if w > mc
    }
    return [mr,mc]
  end

  module_function(:arrange_array, :convert_for_csv, :hash_structize,
    :abs_rc, :rel_rc,
    :eWidth, :eHeight, :eBorder, :eBox, :eAlignment, :eFont,
    :eInterior, :eSpan, :eMerge, :eNumberFormat, :eFormula,
    :eType, :eData, :eHidden, :ePhoneticText,
    :eHRef, :eNamedCell, :getData,
    :eBorder_t, :eBorder_b, :eBorder_l, :eBorder_r, :max_rc)
end


# style򰷤Υ饹
class ExlStyle
  include ExlPrivate

  def initialize(rec = [])
    @record = rec  # Alignment, Borders, Font, Interior,
        # NumberFormat, Protection ξ򤳤˵Ͽ
    @xmlsrc = ''  # style˴ؤxml򤳤˵Ͽ
  end

  def get_id(hs)
    n = @record.index(hs)
    unless n
      @record << hs
      n = @record.size - 1
    end
    return n
  end

  # attr(Hash)饹ʳHash֤
  def set_style(attr)
    hs = Hash.new
    res = Hash.new
    if !attr or attr.size < 1
      return res
    end
    attr.each {|k, v|
      case k
      when 'NumberFormat'
        hs[k] = {'Format'=>arrange_char(v)}
      when 'Font'
        fnt = hash_copy(v)
        if fnt['FontName']
          fnt['FontName'] = arrange_char(fnt['FontName'])
        end
        hs[k] = fnt
      when 'Interior', 'Alignment', 'Borders', 'Protection'
        hs[k] = v
      else
        res[k] = v
      end
    }
    if hs.size > 0
      n = get_id(hs)
      res['StyleID'] = "sid#{n}"
    end
    return res
  end

  def get_xml
    @xmlsrc = Outset_styles.dup
    for n in 0...@record.size
      w = hash_to_xml(@record[n])
      @xmlsrc += tag_attr('Style', {'ID'=>"sid#{n}"}, w)
    end
    @xmlsrc += "</Styles>\n"
    return @xmlsrc
  end
end

# xmlsscell򰷤Υ饹<Cell>ġ</Cell>ν񤭽ФŪ
class ExlTempCell
  include ExlPrivate

  def initialize()
    @index = @may_index = nil
    @chs = Hash.new
  end

  def set(data, index, may_idx)
    @index = index
    @may_index = may_idx
    @chs = Hash.new
    if data.class == Hash
      data.each {|key, val|
        next  unless val
        next  if val.class == Hash and val.size == 0
        @chs[key] = hash_copy(val)
      }
    else
      @chs['Data'] = data.to_s
    end
    append_default_hash(@chs)
  end

  def zen_to_han(str)
    if $ExlapLang == :JAPANESE and str =~ /[-]/
      str = str.gsub(//, "0")
      str.gsub!(//, "1")
      str.gsub!(//, "2")
      str.gsub!(//, "3")
      str.gsub!(//, "4")
      str.gsub!(//, "5")
      str.gsub!(//, "6")
      str.gsub!(//, "7")
      str.gsub!(//, "8")
      str.gsub!(//, "9")
      str.gsub!(//, ",")
      str.gsub!(//, ".")
      str.gsub!(//, "%")
      str.gsub!(//, "/")
      str.gsub!(//, "-")
      str.gsub!(//, ":")
    end
    return str
  end

  def get_number_format(data_str)
    fmt = nil
    ii = ff = nil  # ii:, ff:
    if data_str =~ /^-?(\d[0-9,]*)\.(\d+)$/  # 
      ii = $1
      ff = $2
    elsif data_str =~ /^-?(\d[0-9,]*)$/  # 
      ii = $1
    end
    if ii
      if ii =~ /^0/
        fmt = ii.gsub(/\d/, "0")
      else
        fmt = ii.gsub(/\d/, "#")
        fmt = fmt.sub(/^-/, "")
        fmt = fmt.sub(/\#$/, "0")
        fmt = fmt.sub(/^\#+0$/, "0")
      end
      if fmt =~ /^[0#,]+([0#],[0#]{3})$/
        fmt = $1
      end
      if ff
        fmt = fmt + '.' + ff.gsub(/\d/, "0")
      end
    end
    return fmt
  end

  def auto_type()
    data=@chs['Data']; type=@chs['Type']; number_format=@chs['NumberFormat']
    if !data or data == '' or type == "String" or type =~ /\@[a-z]+$/i
      return
    end
    if data =~ /\n/ or (data_str = zen_to_han(data)) !~ /\d/
      @chs['Type'] = "String"
      return
    end
    if data_str =~ /^-?\d[0-9,]*$|^-?\d[0-9,]*\.\d+$/  # Number
      @chs['Data'] = data_str.gsub(/,/, "").sub(/^(-?)0+/, "\\1")
      @chs['Type'] = "Number"
      fmt = get_number_format(data_str)
      if !number_format and fmt
        @chs['NumberFormat'] = fmt
      end
    elsif data_str =~ /^-?\d[0-9,.]*\%$/  # Percent
      if (fmt = get_number_format(data_str.sub(/\%$/, ""))) != nil
        val = data_str.gsub(/[,%]/, "")
        val = val.to_f / 100.0
        @chs['Data'] = val.to_s
        @chs['Type'] = "Number"
        @chs['NumberFormat'] = fmt + '%'
      end
    elsif data_str =~ /^(\d+)\:(\d+)(\:\d+)?$/
      hh = $1.to_i;  mm = $2.to_i;  ss = $3
      if ss
        ss = ss.sub(/^\:/, "")
        ss = ss.to_i
        @chs['Type'] = "DateTime"
        @chs['NumberFormat'] = "h:mm:ss"
        @chs['Data'] = sprintf("1899-12-31T%02d:%02d:%02d.000", hh, mm, ss)
      else
        @chs['Type'] = "DateTime"
        @chs['NumberFormat'] = "Short Time"
        @chs['Data'] = sprintf("1899-12-31T%02d:%02d:00.000", hh, mm)
      end
    elsif $ExlapLang == :JAPANESE
      if data_str =~ /^(\d{4})[\-\/](\d\d?)[\-\/](\d\d?)$/  # Date
        y = $1.to_i;  m = $2.to_i;  d = $3.to_i
        if y >= 1900 and m >= 1 and m <= 12 and d >= 1 and d <= MaxDay[m]
          @chs['Data'] = sprintf("%d-%02d-%02dT00:00:00.000", y, m, d)
          @chs['Type'] = "DateTime"
          @chs['NumberFormat'] = "Short Date"
        end
      elsif data_str =~ /^(\d{4})ǯ(\d\d?)(\d\d?)$/
        y = $1.to_i;  m = $2.to_i;  d = $3.to_i
        if y >= 1900 and m >= 1 and m <= 12 and d >= 1 and d <= MaxDay[m]
          @chs['Data'] = sprintf("%d-%02d-%02dT00:00:00.000", y, m, d)
          @chs['Type'] = "DateTime"
          @chs['NumberFormat'] = "Long Date"
        end
      elsif data_str =~ /^(|||ʿ)(\d\d?)ǯ(\d\d?)(\d\d?)$/
        gengou = $1
        y = $2.to_i;  m = $3.to_i;  d = $4.to_i
        y += Wareki[gengou]
        if y >= 1868 and m >= 1 and m <= 12 and d >= 1 and d <= MaxDay[m]
          @chs['Data'] = sprintf("%d-%02d-%02dT00:00:00.000", y, m, d)
          @chs['Type'] = "DateTime"
          @chs['NumberFormat'] = '[JPN][$-411]gggy"ǯ"m""d""'
        end
      end
    elsif $ExlapLang == :ENGLISH
      if data_str =~ /^(\d\d?)[\-\/](\d\d?)[\-\/](\d{4})$/  # Date
        m = $1.to_i;  d = $2.to_i;  y = $3.to_i
        if y >= 1900 and m >= 1 and m <= 12 and d >= 1 and d <= MaxDay[m]
          @chs['Data'] = sprintf("%d-%02d-%02dT00:00:00.000", y, m, d)
          @chs['Type'] = "DateTime"
          @chs['NumberFormat'] = "Short Date"
        end
      elsif data_str =~ /^([A-Z][a-z]+)\.? +(\d\d?), +(\d{4})$/
        ms = $1;  d = $2.to_i;  y = $3.to_i
        mn = Month.include?(ms[0,3])
        m = mn ? (mn+1) : 0
        if y >= 1900 and m >= 1 and m <= 12 and d >= 1 and d <= MaxDay[m]
          @chs['Data'] = sprintf("%d-%02d-%02dT00:00:00.000", y, m, d)
          @chs['Type'] = "DateTime"
          @chs['NumberFormat'] = "Long Date"
        end
      end
    end
    if !@chs['Type']
      @chs['Type'] = "String"
    end
  end

  def get_xml(style_info)
    xmlsrc = ''
    if !@index
      return xmlsrc
    end
    cell_attr = Hash.new
    cell_attr['Index'] = @index.to_s  if @index != @may_index
    inner = ''
    @chs.each {|key, val|
      case key
          # for attribute (<Cell ... >)
      when 'Rspan'
        v = val.to_i
        cell_attr['MergeDown'] = (v - 1).to_s  if v > 1
      when 'Cspan'
        v = val.to_i
        cell_attr['MergeAcross'] = (v - 1).to_s  if v > 1
      when 'Formula'
        cell_attr['Formula'] = arrange_char(val)
      when 'HRef'
        cell_attr['HRef'] = arrange_char(val)
      when 'Type'
        if @chs.member?('Data')
          data = @chs['Data'].to_s.dup
          if val !~ /\@[a-z]+$/i
            inner += tag_attr('Data', {'Type'=>val}, arrange_char(data))
          elsif val =~ /\@html$/i
            whs = Hash.new
            whs['Type'] = val.sub(/\@[a-z]+$/i, "")
            whs['xmlns'] = XmlnsHtml
            inner += tag_attr('ss:Data', whs, xnkf(data))
          end
        end
      when 'PhoneticText'
        whs = {'xmlns'=>XmlnsExcel}
        if val.class == Hash
          wtext = val['Text'].to_s
          whs['x:Visible'] = val['Visible'].to_i.to_s
        else
          wtext = val.to_s
        end
        if wtext != ''
          inner += tag_attr('PhoneticText', whs, arrange_char(wtext))
        end
      when 'NamedCell'
        if (w = val.to_s) != ''
          inner += tag_attr('NamedCell', {'Name'=>arrange_char(w)})
        end
      end
    }
    w = style_info.set_style(@chs)
    cell_attr['StyleID'] = w['StyleID']  if w['StyleID']
    xmlsrc += tag_attr('Cell', cell_attr, inner)
    return xmlsrc
  end
end

# Range򰷤Υ饹
class ExlRange
  include ExlPrivate

  def initialize(ss, y1,x1, *arg)
    @sheet = @sy = @sx = @ly = @lx = nil
    if ss.class == ExlSheet and int_t(y1) and int_t(x1)
      case arg.size
      when 0
        @sheet = ss
        @sy, @sx, @ly, @lx = [y1,x1, y1,x1]
      when 2
        if int_t(arg[0]) and
            int_t(arg[1]) and
            arg[0] >= y1 and arg[1] >= x1
          @sheet = ss
          @sy, @sx, @ly, @lx = [y1,x1, arg[0], arg[1]]
        end
      end
    end
    if !@sheet
      arg_str = sprintf("%s,%s,%s,%s", ss, y1, x1, arg.join(','))
      arg_str.sub!(/,""$/, "")
      STDERR.puts "Range can't be initialized (#{arg_str})."
    else
      if (addr = @sheet.merge_address(@sy, @sx))
        @sy, @sx = addr
      end
      if (addr = @sheet.merge_address(@ly, @lx))
        @ly, @lx = [addr[2], addr[3]]
      end
    end
  end
  attr_reader :sheet
  attr_reader :sy
  attr_reader :sx
  attr_reader :ly
  attr_reader :lx

  def in_range(y, x)
    return (y >= @sy and x >= @sx and y <= @ly and x <= @lx) ? true : false
  end
  private :in_range

  def [](y,x)
    if in_range(y+@sy, x+@sx)
      return @sheet[y+@sy, x+@sx]
    end
    return false
  end

  def []=(y, x, val)
    if in_range(y+@sy, x+@sx)
      y += @sy
      x += @sx
      @sheet[y,x] = val
      return @sheet[y,x]
    end
    return false
  end

  def Data=(val)
    mxy = max_y()
    mxx = max_x()
    if val.class == Array and val[0].class == Array and  # ξ
        val[0][0].class != Array
      val.each_with_index {|row, y|
        break  if y > mxy
        row.each_with_index {|data, x|
          break  if x > mxx
          if data != nil
            self.cell(y,x).Data = data
          else
            self.cell(y,x).delete('Data')
          end
        }
      }
    else  # ξ
      val = (val.class != Array) ? [val] : val.dup
      for y in 0..mxy
        break  if val.size == 0
        for x in 0..mxx
          break  if val.size == 0
          if (w = val.shift) != nil
            self.cell(y,x).Data = w
          else
            self.cell(y,x).delete('Data')
          end
        end
      end
    end
    return self
  end

  def Data()
    ary = []
    for y in @sy..@ly
      row = []
      for x in @sx..@lx
        row << getData(@sheet[y,x])
      end
      ary << row
    end
    return ary
  end

  def max_row()
    return (@ly - @sy + 1)
  end

  def max_col()
    return (@lx - @sx + 1)
  end

  def max_y()
    return (@ly - @sy)
  end

  def max_x()
    return (@lx - @sx)
  end

  def cell(y,x)
    if in_range(y+@sy, x+@sx)
      return @sheet.cell(y+@sy, x+@sx)
    end
    return false
  end

  def range(y, x, *arg)
    y1 = y + @sy
    x1 = x + @sx
    if arg.size == 2
      y2 = arg[0] + @sy
      x2 = arg[1] + @sx
    elsif arg.size == 0
      y2,x2 = [y1,x1]
    else
      y2,x2 = [-1, -1]
    end
    if in_range(y1,x1) and in_range(y2,x2)
      return ExlRange.new(@sheet, y1,x1, *arg)
    end
    return false
  end

  def address(*arg)
    if arg.size == 0
      return [@sy, @sx, @ly, @lx]
    elsif arg.size == 2
      y,x = arg
      return in_range(y+@sy, x+@sx) ? [y+@sy, x+@sx] : nil
    end
    return nil
  end

  def row_range(y)
    y += @sy
    return (y>=@sy and y<=@ly) ? @sheet.range(y,@sx, y,@lx) : false
  end

  def col_range(x)
    x += @sx
    return (x>=@sx and x<=@lx) ? @sheet.range(@sy,x, @ly,x) : false
  end

  def each()
    unless block_given?
      return self
    end
    span = merge_span()
    mr,mc = Exl::max_rc(span)
    for y in 0...mr
      for x in 0...mc
        next  if span[y][x]
        yield(self.cell(y,x))
      end
    end
    return self
  end

  def each_with_yx()
    unless block_given?
      return self
    end
    span = merge_span()
    mr,mc = Exl::max_rc(span)
    for y in 0...mr
      for x in 0...mc
        next  if span[y][x]
        yield(self.cell(y,x), y, x)
      end
    end
    return self
  end

  def clear()
    mxy, mxx = [self.max_y, self.max_x]
    for y in 0..mxy
      for x in 0..mxx
        self[y,x] = nil
      end
    end
    return self
  end

  def merge_cells=(val)
    unless [true,false].include?(val)
      return false
    end
    if val == false
      cl = cell(0,0)
      cl.delete('Rspan')  if cl.member?('Rspan')
      cl.delete('Cspan')  if cl.member?('Cspan')
      return self
    end
    for y in 0...max_row()
      for x in 0...max_col()
        cl = cell(y,x)
        if y == 0 and x == 0
          cl.delete('Rspan')  if cl.member?('Rspan')
          cl.delete('Cspan')  if cl.member?('Cspan')
          cl['Rspan'] = max_row()  if max_row() > 1
          cl['Cspan'] = max_col()  if max_col() > 1
        else
          cell(0,0).mrg cl['Borders']  if cl.member?('Borders')
          cl.clear
        end
      end
    end
    return self
  end

  def merge_span()
    ary = []
    max_r = @ly-@sy+1
    max_c = @lx-@sx+1
    max_r.times do
      ary << []
    end
    for y in 0...max_r
      for x in 0...max_c
        next  if ary[y][x]
        ary[y][x] = nil
        cl = self[y,x]
        next  unless cl.class == Hash
        rspan = cl.member?('Rspan') ? cl['Rspan'].to_i : 1
        cspan = cl.member?('Cspan') ? cl['Cspan'].to_i : 1
        next  if rspan <= 1 and cspan <= 1
        for rn in y...(y+rspan)
          ary[rn] = []  unless ary[rn]
          xx = (rn == y) ? (x+1) : x
          for cn in xx...(x+cspan)
            ary[rn][cn] = [y,x]
          end
        end
      end
    end
    return ary
  end

  def clearbox()
    if @sheet
      self.each {|cl|
        cl.delete('Borders')  if cl.member?('Borders')
      }
    end
    return self
  end

  def allbox(line_style=nil, weight=nil, color=nil)
    if @sheet
      self.each {|cl|
        cl.mrg(eBorder(line_style, weight, color))
      }
    end
    return self
  end

  def framebox(line_style=nil, weight=nil, color=nil)
    if @sheet and !@sheet.empty?
      span = merge_span()
      mr, mc = Exl::max_rc(span)
      for y in 0...mr
        next  unless span[y]
        for x in 0...mc
          cl = (w = span[y][x]) ? cell(*w) : cell(y,x)
          cl.mrg eBorder_t(line_style, weight, color)  if y == 0
          cl.mrg eBorder_b(line_style, weight, color)  if y == mr-1
          cl.mrg eBorder_l(line_style, weight, color)  if x == 0
          cl.mrg eBorder_r(line_style, weight, color)  if x == mc-1
        end
      end
    end
    return self
  end
end

# ĤΥȤ򰷤Υ饹
class ExlSheet
  include ExlPrivate

  def initialize(ary=nil, sn=nil, src=nil)
    @array = nil  # all cells of the worksheet(Array of Array)
    if ary.class == Array and ary[0].class == Array and
       ary[0][0].class != Array
      @array = ary
    end
    set_name(sn)
    @start_rc = src  # start position in the worksheet
    @tbl = {}
    @row = []  # rows' infomation of the worksheet(Array)
    @col = []  # columns' infomation of the worksheet(Array)
    @source_type = nil  # the source type of array (:WEB | :CSV)
    @source_clue = nil  # the source clue (url | csv_filename | csv_string)
    @other = Hash.new
  end
  attr_accessor :start_rc
  attr_accessor :tbl
  attr_accessor :row
  attr_accessor :col
  attr_accessor :source_type
  attr_accessor :source_clue
  attr_accessor  :other

  def empty?()
    res = false
    if !@array or @array == [] or @array == [[]]
      res = true
    end
    return res
  end

  def array=(ary)
    if ary.class == Array and ary[0].class == Array and
       ary[0][0].class != Array
      @array = ary
    end
  end

  def array()
    return @array
  end

  def clear()
    @array = nil
  end

  def [](y,x)
    if self.empty? or !@array[y]
      return nil
    end
    return @array[y][x]
  end

  def []=(y,x, val)
    unless @array
      @array = []
    end
    unless @array[y]
      for wy in max_row()..y
        @array << []
      end
    end
    if val.class == Hash
      @array[y][x] = Exl::hash_structize(val)
    else
      @array[y][x] = val
    end
    return @array[y][x]
  end

  def cell(y,x)
    val = self[y,x]
    if val.class == Hash
      self[y,x] = Exl::hash_structize(val)
    elsif val == nil
      self[y,x] = {}.structize
    else
      self[y,x] = {'Data'=>val}.structize
    end
    return @array[y][x]
  end

  def set_name(sn)
    return  unless sn
    w = xnkf(sn.to_s)
    ptn = (RUBY_VERSION < "1.9") ?
        Regexp.new('[Ρ]', nil,
            ($ExlapBaseC == 'w') ? 'u' : $ExlapBaseC) : /[Ρ]/
    if sn =~ /[:?*\\\/\[\]]/ or w =~ ptn
      STDERR.puts "the sheet name includes unavailable char: '#{sn}'"
    elsif str_len(sn) > 31
      STDERR.puts "the sheet name is too long: '#{sn}'"
    else
      @name = sn
    end
  end

  def name=(sn)
    set_name(sn)
  end

  def name()
    return @name
  end

  def max_col()  # ĹιԤ֤
    mr,mc = Exl::max_rc(@array)
    return mc
  end

  def max_row()
    return @array ? @array.size : 0
  end

  def array_update(hs)
    if self.empty? or hs.class != Hash
      return self
    end
    code = array_code(@array)
    if code
      hs.each {|key, val|
        case key
        when 'Data'
          hs[key] = xnkf(val, code)  if val.class == String
        when 'NumberFormat'
          hs[key] = xnkf(val, code)  if val.class == String
        when 'Font'
          if (font_name = hs[key]['FontName'])
            hs[key]['FontName'] = xnkf(font_name, code)
          end
        end
      }
    end
    for rn in 0...@array.size
      next  if !@array[rn] or @array[rn].size < 1
      for cn in 0...@array[rn].size
        next  if !@array[rn][cn]
        if @array[rn][cn].class == Hash
          @array[rn][cn] = Exl::hash_structize(@array[rn][cn])
          @array[rn][cn].mrg(hs)
        else
          @array[rn][cn] = {'Data'=>@array[rn][cn].to_s}.structize
          @array[rn][cn].mrg(hs)
        end
      end
    end
    return self
  end

  def array_convert(scr)
    if self.empty? or scr.class != String
      return self
    end
    code = array_code(@array)
    if code
      scr = xnkf(scr, code)
    end
    for rn in 0...@array.size
      next  if !@array[rn] or @array[rn].size < 1
      for cn in 0...@array[rn].size
        next  if !@array[rn][cn]
        if @array[rn][cn].class == Hash
          str = @array[rn][cn]['Data']
          if str
            str = str.to_s
            eval(scr)
            @array[rn][cn]['Data'] = str
          end
        else
          str = @array[rn][cn].to_s
          eval(scr)
          @array[rn][cn] = str
        end
      end
    end
    return self
  end

  # ĤΥȤxmlssscriptѴ
  def get_xml(style_info)
    if !@name or @name == ''
      sheet_name = "Sheet #{Time.now.strftime('%Y-%m-%d %H_%M_%S')}" +
          "r#{rand(1000)}"
    else
      sheet_name = arrange_char(@name.to_s)
    end
    if @start_rc and @start_rc.class == Array and @start_rc.size == 2
      start_r = @start_rc[0].to_i
      start_c = @start_rc[1].to_i
    else
      start_r = start_c = 0
    end
    max_r = max_row()
    max_c = max_col()
    xmlsrc = ''
    if max_r > 0 and max_c > 0  # ȤǤϤʤ
      cs = ExlTempCell.new
      colsrc = ''
      c_idx = 1
      for i in 0...@col.size
        next  if !@col[i] or @col[i].class != Hash
        w = hash_copy(@col[i])
        if !w['Column']
          w = {'Column'=>w}
        end
        append_default_hash(w['Column'])
        n = i+1+start_c
        if n != c_idx
          w['Column'].update({'Index'=>"#{n}"})
        end
        c_idx = (n+1)
        w['Column'] = style_info.set_style(w['Column'])
        colsrc += hash_to_xml(w)
      end
      aa = arrange_array_s(@array)
      r_idx = 1
      for rn in 0...max_r
        next  if !aa[rn] or aa[rn].size < 1
        col_str = ''
        c_idx = 1
        for cn in 0...aa[rn].size
          next  unless aa[rn][cn]
          data = aa[rn][cn]
          if data.class != Hash
            data = data.to_s
          else
            if data.size == 0
              next
            elsif data.member?('Data')
              data['Data'] = data['Data'].to_s
            end
          end
          cs.set(data, cn+1+start_c, c_idx)
          cs.auto_type
          col_str += cs.get_xml(style_info)
          c_idx = cn+1+start_c + 1
        end
        next  if col_str == ''
        nn = rn+1+start_r
        if !@row[rn] or @row[rn].class != Hash
          if nn != r_idx
            row_attr = {'Index'=>"#{nn}"}
          else
            row_attr = {}
          end
        else
          row_attr = hash_copy(@row[rn])
          if row_attr['Row']
            row_attr = row_attr['Row']
          end
          if nn != r_idx
            row_attr.update({'Index'=>"#{nn}"})
          end
        end
        append_default_hash(row_attr)
        row_attr = style_info.set_style(row_attr)
        xmlsrc += tag_attr('Row', row_attr, col_str)
        r_idx = nn + 1
      end
      if xmlsrc != ''
        outset = Outset_worksheet.dup
        outset.sub!(/__sheet_name__/, sheet_name)
        tbl = hash_copy(@tbl)
        append_default_hash(tbl)
        tbl = Table_attr.dup.update(tbl)
        tbl['ExpandedRowCount'] = (max_r + start_r).to_s
        tbl['ExpandedColumnCount'] = (max_c + start_c).to_s
        tbl = style_info.set_style(tbl)
        w = ''
        @other.each {|k,v|
          w = w + xnkf(v)
        }
        xmlsrc = colsrc + xmlsrc
        xmlsrc = outset + tag_attr('Table', tbl, xmlsrc) + w + "</Worksheet>"
      end
    end
    if xmlsrc == ''
      xmlsrc = "<Worksheet ss:Name=\"#{sheet_name}\">" +
        "<Table ss:ExpandedColumnCount=\"1\" ss:ExpandedRowCount=\"1\" " +
        "x:FullColumns=\"1\" x:FullRows=\"1\" ss:DefaultColumnWidth=\"54\" " +
        "ss:DefaultRowHeight=\"13.5\"></Table></Worksheet>\n"
    end
    return xmlsrc
  end

  def length_of_col(obj)
    data = nil
    cspan = 1
    if obj.class == Hash
      data = obj['Data']
      w = obj['Cspan']
      cspan = w.to_i  if w
    else
      data = obj
    end
    if !data
      return 0
    end
    s = data.to_s.gsub(/[^\x00-\x7f]/, "!x")
    s = s.gsub(/\t/, (' ' * $ExlapTabLen))
    len = s.length / cspan
    if s =~ /\n/
      w = 0
      s.split(/\n/).each {|line|
        l = line.length
        w = l  if l > w
      }
      len = w / cspan
    end
    return len
  end
  alias :loc :length_of_col

  def max_length_of_col(rs=nil, re=nil)
    rs = 0  if !rs
    re = (max_row() - 1)  if !re
    max_c = max_col()
    if re < 0 or max_c < 1
      return nil
    end
    res = Array.new(max_c, 0)
    code = 's'
    code = $ExlapBaseC  if ['e', 's', 'w'].include?($ExlapBaseC)
    aa = arrange_array_s(@array, code)
    change_kcode(code)
    for rn in 0...aa.size
      next  if rn < rs
      break  if rn > re
      next  if !aa[rn] or aa[rn].size < 1
      for cn in 0...aa[rn].size
        next  if !aa[rn][cn]
        len = loc(aa[rn][cn])
        res[cn] = len  if len > res[cn]
      end
    end
    change_kcode()
    return res
  end
  alias :mloc :max_length_of_col

  # ƥСʸɤĴ
  def all_code_adj()
    @array = arrange_array_s(@array)
    @name = code_adj(@name)
    @tbl = code_adj(@tbl)
    @row = code_adj(@row)
    @col = code_adj(@col)
    @source_clue = code_adj(@source_clue)
    @other = code_adj(@other)
  end

  def range(y1,x1, *arg)
    return ExlRange.new(self, y1,x1, *arg)
  end

  def entire_range()
    res = false
    if self.empty?
      return res
    end
    span = range(0,0, max_row()-1, max_col()-1).merge_span()
    mr,mc = Exl::max_rc(span)
    res = ExlRange.new(self, 0,0, mr-1, mc-1)
  end

  def used_range()
    if self.empty?
      return false
    end
    span = range(0,0, max_row()-1, max_col()-1).merge_span()
    mr,mc = Exl::max_rc(span)
    y2,x2 = [mr-1, mc-1]
    for y in 0..y2
      empty_row = true
      for x in 0..x2
        if !span[y][x] and self[y,x] and self[y,x] != {}
          empty_row = false
          break
        end
      end
      break  unless empty_row
    end
    y1 = empty_row ? 0 : y
    for x in 0..x2
      empty_col = true
      for y in y1..y2
        if !span[y][x] and self[y,x] and self[y,x] != {}
          empty_col = false
          break
        end
      end
      break  unless empty_col
    end
    x1 = empty_col ? 0 : x
    return ExlRange.new(self, y1,x1, y2,x2)
  end

  def row_range(y)
    res = false
    max_r = max_row()
    if max_r > 0 and y < max_r and @array[y].size > 0
      res = self.range(y,0, y,@array[y].size-1)
    end
    return res
  end

  def col_range(x)
    res = false
    max_r = max_row()
    max_c = max_col()
    if max_r > 0 and max_c > 0 and x < max_c
      res = self.range(0,x, max_r-1,x)
    end
    return res
  end

  def copy_from(*arg)
    if arg.size == 3 and
          ((cl = arg[2].class) == ExlRange or
          cl == ExlSheet and !self.equal?(arg[2]))
      y,x,obj = arg
      other = (cl == ExlRange) ? obj :
          obj.range(0,0, obj.max_row-1, obj.max_col-1)
      rng = self.range(y,x, y+other.max_y, x+other.max_x)
    elsif arg.size == 1 and
          ((cl = arg[0].class) == ExlRange or
          cl == ExlSheet and !self.equal?(arg[0]))
      obj = arg[0]
      other = (cl == ExlRange) ? obj :
          obj.range(0,0, obj.max_row-1, obj.max_col-1)
      rng = self.range(0,0, other.max_y, other.max_x)
    else
      return false
    end
    rng.each_with_yx {|c, y, x|
      c.clear
      c.update other.cell(y,x)
    }
    return rng
  end

  def merge_address(yy, xx)
    res = nil
    if self.empty?
      return res
    end
    yy.step(0, -1) {|y|
      next  if !@array[y]
      xx.step(0, -1) {|x|
        cl = self[y,x]
        next  if cl.class != Hash or cl == {}
        rspan = cl['Rspan'] ? (cl['Rspan'].to_i-1) : 0
        cspan = cl['Cspan'] ? (cl['Cspan'].to_i-1) : 0
        next  if rspan < 1 and cspan < 1
        if yy>=y and yy<=(y+rspan) and xx>=x and xx<=(x+cspan)
          res = [y, x, y+rspan, x+cspan]
          break
        end
      }
      break  if res
    }
    return res
  end

  def merge_area(yy, xx)
    addr = merge_address(yy, xx)
    rng = addr ? self.range(*addr) : false
  end
end


# ֥å򰷤δܥ饹
class ExlBookBase
  include ExlPrivate

  def initialize
    @sheets = Array.new
    @author = @last_author = @company = nil
    @created = @last_saved = nil
    @styles = ExlStyle.new
    @other = Hash.new
    @other_dp = Hash.new
  end
  attr_accessor  :sheets
  attr_accessor  :author
  attr_accessor  :last_author
  attr_accessor  :company
  attr_accessor  :created
  attr_accessor  :last_saved
  attr_accessor  :other

  # Ⱦ򥻥å(ʣ)
  def set_sheet(*params)
    while aa = params.shift
      if aa.class == ExlSheet
        @sheets << aa
        next
      end
      next  if aa.class != Array
      next  if aa[0].class != Array
      sheet_name = nil
      start_rc = nil
      while params.size > 0
        if params[0].class == String
          sheet_name = params.shift
        elsif params[0].class == Array and params[0][0].class != Array
          start_rc = params.shift
        else
          break
        end
      end
      esh = ExlSheet.new(aa, sheet_name, start_rc)
      @sheets << esh
    end
    return self
  end

  # Ⱦˤ xml
  def get_xml()
    xmlsrc = ''
    snames = Array.new
    if !@sheets or @sheets.size < 1
      return xmlsrc
    end
    ary = Array.new
    for i in 0...@sheets.size
      ary[i] = @sheets[i].dup
    end
    for i in 0...ary.size
      if ary[i].class == ExlSheet
        ary[i].name = arrange_sheet_name(ary[i].name, snames, i)
        xmlsrc += ary[i].get_xml(@styles)
      else
        next  if ary[i].class != Array
        next  if ary[i][0].class != Array
        ary[i] = ExlSheet.new(ary[i], sprintf("Sheet%d", i+1))
        xmlsrc += ary[i].get_xml(@styles)
      end
    end
    unless @created
      @created = "#{Time.now.gmtime.strftime('%Y-%m-%dT%H:%M:%SZ')}"
    end
    @last_saved = "#{Time.now.gmtime.strftime('%Y-%m-%dT%H:%M:%SZ')}"
    outset = Outset_workbook.dup
    outset.gsub!(/__created__/, @created)
    outset.gsub!(/__last_saved__/, @last_saved)
    outset.sub!(/__author__/, arrange_char(@author))  if @author
    outset.sub!(/__last_author__/, arrange_char(@last_author))  if @last_author
    outset.sub!(/__company__/, arrange_char(@company))  if @company
    if @other_dp.size > 0
      w = ''
      @other_dp.each {|k,v|
        w = w + xnkf(v)
      }
      outset.sub!(/<\/DocumentProperties>/, w + "\\&")
    end
    w = ''
    @other.each {|k,v|
      w = w + xnkf(v)
    }
    xmlsrc = outset + @styles.get_xml + w + xmlsrc + "</Workbook>"
    xmlsrc = xnkf(xmlsrc, 'w')  unless $ExlapBaseC == 'w'
    return xmlsrc
  end

  def xml_string
    if @sheets.size < 1
      return ''
    end
    return get_xml()
  end

  def output_xml(filename)
    str = xml_string()
    if str != ''
      File.open(filename, "w") {|ff|  ff.print str}
    end
  end
  alias :xml_output :output_xml

  def output_csv(filename)
    if @sheets.size < 1
      return
    end
    csv_ary = arrange_for_csv(*@sheets)
    if csv_ary.size < 1
      return
    end
    if RUBY_VERSION < "1.9" and defined?(FasterCSV)
      FasterCSV.open(filename, "w") {|writer|
        csv_ary.each {|aa|
          aa.each {|row|
            writer << (!row ? [] : row)
          }
          writer << []
          writer << []
        }
      }
    else
      CSV.open(filename, "w") {|writer|
        csv_ary.each {|aa|
          aa.each {|row|
            writer << (!row ? [] : row)
          }
          writer << []
          writer << []
        }
      }
    end
  end
  alias :csv_output :output_csv

  def web_array()
    res = Array.new
    if !@sheets or @sheets.size < 1
      return res
    end
    snames = Array.new
    @sheets.each_with_index {|ss, count|
      sheet_name = arrange_sheet_name(arrange_char(
          recover_char(xnkf(ss.name.to_s))), snames, count)
      aa = arrange_array_s(ss, nil, :TO_STRING)
      span = {}
      tbl_str = "<h2 id=\"ExlSheet#{count+1}\">#{sheet_name}</h2>\n" +
          "<table id=\"ExlSheet#{count+1}\">"
      for rn in 0...aa.size
        if rn == 0 and aa[rn] =~ /^caption\:(.+)$/i
          w = $1
          tbl_str += "<caption>#{arrange_char(w)}</caption>\n"
          next
        end
        tr_str = '<tr>'
        for cn in 0...aa[rn].size
          next  if span[[rn,cn]]
          cell_str = arrange_char(aa[rn][cn].to_s)
          cell = ss.array[rn][cn]
          if cell.class == Hash and (cell['Cspan'] or cell['Rspan'])
            cspan = cell['Cspan'] ? cell['Cspan'].to_i : 1
            rspan = cell['Rspan'] ? cell['Rspan'].to_i : 1
            if cspan > 1 or rspan > 1
              for r in rn...(rn+rspan)
                for c in cn...(cn+cspan)
                  next  if r == rn and c == cn
                  span[[r,c]] = true
                end
              end
              td_attr = ''
              td_attr += " colspan=\"#{cspan}\""  if cspan > 1
              td_attr += " rowspan=\"#{rspan}\""  if rspan > 1
              tr_str += "<td id=\"#{cn+1}\"#{td_attr}>#{cell_str}</td>"
            else
              tr_str += (cell_str != '') ?
                "<td id=\"#{cn+1}\">#{cell_str}</td>" : "<td id=\"#{cn+1}\"/>"
            end
          else
            tr_str += (cell_str != '') ?
                "<td id=\"#{cn+1}\">#{cell_str}</td>" : "<td id=\"#{cn+1}\"/>"
          end
        end
        tr_str += "</tr>\n"
        tr_str = tr_str.sub(/<tr><\/tr>/, "<tr/>")
        tbl_str += tr_str
      end
      tbl_str = tbl_str.gsub(/\&\#13;\&\#10;/, "<br/>")
      while tbl_str =~ /<tr\/>\n<tr\/>/
        tbl_str = tbl_str.gsub(/<tr\/>\n<tr\/>/, "<tr/><tr/>")
      end
      tbl_str += "</table>\n"
      res << tbl_str
    }
    return res
  end

  def web_string(title = nil)
    ary = web_array()
    if ary.size < 1
      return ''
    end
    body_str = ''
    if ary.size > 1
      contents_str = "<div id=\"contents\">\n<h2>contents</h2><ul>\n"
      ary.each_with_index {|tbl, count|
        /<h2[^>]*>(.+?)<\/h2>/ =~ tbl
        sheet_name = Regexp.last_match(1)
        contents_str += sprintf("<li><a href=\"#sn%d\">%s</a></li>\n",
            count+1, sheet_name)
        body_str += "<a name=\"sn#{count+1}\"></a>\n#{tbl}<hr/>\n\n"
      }
      contents_str += "</ul>\n<hr/>\n</div>\n\n"
      body_str = contents_str + body_str
    else
      body_str = ary[0]
    end
    res = Outset_html.dup
    res.sub!(/__title__/, title ? title : 'no title')
    case $ExlapBaseC
    when 'w'
      jcode = 'utf-8'
    when 's'
      jcode = 'Shift_JIS'
    when 'e'
      jcode = 'euc-jp'
    else
      jcode = 'utf-8'
    end
    res.sub!(/__jcode__/, jcode)
    res = res + body_str + "</body></html>"
    return res
  end

  def output_web(filename, title=nil)
    str = web_string(title)
    if str != ''
      File.open(filename, "w") {|ff|  ff.print str}
    end
  end
  alias :web_output :output_web
end


# Webǡ뤿Υ饹
  # MS-WindowsξMsxml2.xmlhttpѡ¾open-uri
class Mshttp
  def initialize(os_val = :COMMON)
    @os_val = os_val
    @MSHTTP = nil
    if @os_val == :WIN32
      require 'win32ole'
      begin
        @MSHTTP = WIN32OLE.new('Msxml2.XMLHTTP')
      rescue
        @MSHTTP = nil
      end
      unless defined? @MSHTTP
        begin
          @MSHTTP = WIN32OLE.new('MSXML.XMLHTTPRequest')
        rescue
          @MSHTTP = nil
        end
      end
      unless defined? @MSHTTP
        begin
          @MSHTTP = WIN32OLE.new('Microsoft.XMLHTTP')
        rescue
          @MSHTTP = nil
        end
      end
      @sec_limit = 20  # ॢÿ
    else
      require 'open-uri'
    end
    @body = ''
  end
  attr_accessor  :sec_limit
  attr_reader  :body

  # htmlɥȤԻ false ޤ '' ֤
  def web_open(url)
    if @os_val == :WIN32
      unless @MSHTTP
        return false
      end
      if url =~ /^[ \t]*http[\x21-\x7e]+/i
        @MSHTTP.Open "GET", url, false
        sec_count = 0
        while @MSHTTP.readyState != 1
          sleep 1
          sec_count += 1
          break  if sec_count > @sec_limit
        end
        if sec_count > @sec_limit
          @MSHTTP.abort
          return false
        end
        begin
          @MSHTTP.Send
        rescue
          @MSHTTP.abort
          return false
        end
        sec_count = 0
        while @MSHTTP.readyState != 4
          sleep 1
          sec_count += 1
          break  if sec_count > @sec_limit
        end
        if sec_count > @sec_limit
          @MSHTTP.abort
          return false
        end
        if @MSHTTP.status == 200
          @body = @MSHTTP.responseBody.pack("c*")
        else
          @body = ''
        end
        @MSHTTP.abort
      end
    else
      begin
        @body = open(url).read
      rescue
        return false
      end
    end
    return @body
  end
end
$ExlapMshttp = Mshttp.new($ExlapOs)


# ֥å򰷤Υ饹
class ExlBook < ExlBookBase
  include ExlPrivate

  def initialize(*params)
    @sheets = Array.new
    @author = @last_author = @company = nil
    @created = @last_saved = nil
    @styles = ExlStyle.new
    @other = Hash.new
    @other_dp = Hash.new
    push(*params)  if params.size > 0
  end

  def web_push(*url_info)
    ew = ExlWeb.new
    ew.push(*url_info)
    set_sheet(*ew.sheets)
    return self
  end
  alias :push_web :web_push

  def csv_push(*opt)
    ec = ExlCsv.new
    ec.push(*opt)
    set_sheet(*ec.sheets)
    return self
  end
  alias :push_csv :csv_push

  def push(*params)
    cl = params[0].class
    if cl == ExlSheet or (cl == Array and params[0][0].class == Array)
      set_sheet(*params)
    elsif cl == String and (params[0] =~ /^[ \t]*http[\x21-\x7e]+$/i or
        is_html(params[0]))
      web_push(*params)
    elsif cl == String
      csv_push(*params)
    end
    return self
  end

  def <<(*params)
    return push(*params)
  end

  def recover_charw(x)
    return recover_char(x, 'w')
  end

  # xmlss
  def xml_push(clue, other_sw=nil)
    str = (clue.class == String) ? clue.dup : clue.to_s
    if str =~ /^[ \t]*http[\x21-\x7e]+/i
      body = $ExlapMshttp.web_open(str)
      if body == '' or body == false
        STDERR.printf("could not access: %s\n", str)
        return self
      end
      body = xnkf(body).gsub(/\r\n/, "\n")
    elsif !is_xmlss((body = xnkf(str)))
      begin
        body = File.open(str) {|ff|  ff.read}
        body = xnkf(body)
      rescue
        STDERR.printf("can't read '%s' as a file\n", str)
        return self
      end
    end
    unless is_xmlss(body)
      STDERR.printf("do not recognize as xmlss\n")
      return self
    end
    change_kcode('u')
    doc = REXML::Document.new(xnkf(body, 'w'))
        ## DocumentProperties
    doc.elements["/Workbook/DocumentProperties"].elements.each {|elem|
      key = elem.expanded_name
      val = elem.get_text.to_s
      case key
      when 'Author'
        @author = recover_charw(val)  if !@author
      when 'LastAuthor'
        @last_author = recover_charw(val)  if !@last_author
      when 'Company'
        @company = recover_charw(val)  if !@company
      when 'Created'
        @created = recover_charw(val)  if !@created
      when 'LastSaved'
        @last_saved = recover_charw(val)  if !@last_saved
      else
        if other_sw and !@other_dp[key]
          @other_dp[key] = xnkf(elem.to_s)
        end
      end
    }
        ## get styles infomation
    styles_rec = Hash.new
    doc.elements["/Workbook/Styles"].elements.each {|elem|
      next  if elem.expanded_name != "Style"
      id = elem.attributes['ss:ID']
      next  if id.downcase == "default"
      styles_rec[id] = {}
      elem.elements.each {|elem2|
        if elem2.expanded_name == "Borders"
          count = 0
          hs = Hash.new
          elem2.elements.each {|elem3|
            next  unless elem3.has_attributes?
            hs2 = Hash.new
            elem3.attributes.each {|key, val|
              key = key.sub(/^ss\:/, "")
              hs2[key] = val
            }
            count += 1
            tagname = "#{elem3.expanded_name}#{count}"
            hs2 = {tagname=>hs2}
            hs.update(hs2)
          }
          styles_rec[id].update({elem2.expanded_name=>hs})
        elsif elem2.expanded_name == "NumberFormat"
          next  unless elem2.has_attributes?
          if (val = elem2.attributes['ss:Format'])
            val = recover_charw(val)
            styles_rec[id].update({elem2.expanded_name=>val})
          end
        else
          next  unless elem2.has_attributes?
          hs = Hash.new
          elem2.attributes.each {|key, val|
            key = key.sub(/^ss\:/, "")
            if ['FontName'].include?(key)
              val = recover_charw(val)
            end
            hs[key] = val
          }
          styles_rec[id].update({elem2.expanded_name=>hs})
        end
      }
    }
    tmp_hs = {}
    styles_rec.each {|key, val|
      wsid = @styles.set_style(val)
      tmp_hs[key] = wsid['StyleID']  if wsid['StyleID']
    }
        ## get worksheet infomation
    sheet_ary = []
    doc.elements["/Workbook"].elements.each {|elem|
      tag_name = elem.expanded_name
      unless ['DocumentProperties', 'Styles',
          'Worksheet'].include?(tag_name)
        if other_sw
          if !@other[tag_name]
            @other[tag_name] = xnkf(elem.to_s)
          else
            @other[tag_name] += xnkf(elem.to_s)
          end
        end
        next
      end
      next  if tag_name != "Worksheet"
      ss = ExlSheet.new
      ss.name = recover_charw(elem.attributes['ss:Name'])
      elem.elements.each {|elem2|
        if (tag_name = elem2.expanded_name) != 'Table'
          if other_sw
            if !ss.other[tag_name]
              ss.other[tag_name] = xnkf(elem2.to_s)
            else
              ss.other[tag_name] += xnkf(elem2.to_s)
            end
          end
        end
      }
      next  unless (te = elem.elements['Table'])
      hs = Hash.new
      te.attributes.each {|key, val|
        key = key.sub(/^ss\:/, "")
        if key == 'StyleID'
          val = tmp_hs[val]
        end
        hs[key] = val
      }
      next  if hs.size < 1
      ss.tbl = hs
      max_c = hs['ExpandedColumnCount'].to_i
      max_r = hs['ExpandedRowCount'].to_i
      next  if max_c < 1 or max_r < 1
      ary = Array.new
      max_r.times do
        ary << Array.new(max_c)
      end
      ss.array = ary
      col_count = row_count = 0
      span = {}
      te.elements.each {|elem2|
        if elem2.expanded_name == 'Column'
          col_count += 1
          next  unless elem2.has_attributes?
          if (idx = elem2.attributes['ss:Index'])
            col_count = idx.to_i
          end
          hs = Hash.new
          elem2.attributes.each {|key, val|
            next  if key == 'ss:Index'
            key = key.sub(/^ss\:/, "")
            if key == 'StyleID'
              val = tmp_hs[val]
            end
            hs[key] = val
          }
          ss.col[col_count-1] = hs  if hs.size > 0
        elsif elem2.expanded_name == 'Row'
          row_count += 1
          if elem2.has_attributes?
            if (idx = elem2.attributes['ss:Index'])
              row_count = idx.to_i
            end
            hs = Hash.new
            elem2.attributes.each {|key, val|
              next  if key == 'ss:Index'
              key = key.sub(/^ss\:/, "")
              if key == 'StyleID'
                val = tmp_hs[val]
              end
              hs[key] = val
            }
            ss.row[row_count-1] = hs  if hs.size > 0
          end
          cell_count = 0
          elem2.elements.each {|elem3|
            next  unless elem3.expanded_name == 'Cell'
            cell_count += 1
            while span[[row_count, cell_count]]
              cell_count += 1
            end
                # for attribute (<Cell ... >)
            hs = Hash.new
            sid = nil
            cspan = rspan = 1
            if elem3.has_attributes?
              elem3.attributes.each {|key2, val2|
                key2 = key2.sub(/^ss\:/, "")
                case key2
                when 'Index'
                  cell_count = val2.to_i
                when 'StyleID'
                  sid = val2
                when 'MergeAcross'
                  cspan = hs['Cspan'] = val2.to_i + 1
                when 'MergeDown'
                  rspan = hs['Rspan'] = val2.to_i + 1
                when 'Formula', 'HRef'
                  hs[key2] = recover_charw(val2)
                end
              }
            end
            if cspan > 1 or rspan > 1
              for r in row_count...(row_count+rspan)
                for c in cell_count...(cell_count+cspan)
                  next  if r == row_count and c == cell_count
                  span[[r,c]] = true
                end
              end
            end
                # for inner (<Cell> ... </Cell>)
            elem3.elements.each {|elem4|
              key3 = elem4.expanded_name
              case key3
              when 'Data'
                if elem4.has_attributes? and (w = elem4.attributes['ss:Type'])
                  hs['Type'] = w
                end
                if (w = elem4.get_text)
                  w = w.to_s
                  hs[key3] = recover_charw(w)  if w != ''
                end
              when 'ss:Data'
                if elem4.has_attributes? and (w = elem4.attributes['ss:Type'])
                  hs['Type'] = w + "@html"
                end
                w = elem4.to_s
                w.sub!(/<ss\:Data [^>]+>/i, "")
                w.sub!(/<\/ss\:Data>/i, "")
                hs['Data'] = w  if w != ''
              when 'PhoneticText'
                wtext = wvisible = nil
                if (w = elem4.get_text)
                  w = w.to_s
                  wtext = recover_charw(w)  if w != ''
                end
                if wtext and elem4.has_attributes? and (wvisible =
                    elem4.attributes['x:Visible'])
                  hs[key3] = {'Text'=>wtext, 'Visible'=>wvisible}
                elsif wtext
                  hs[key3] = wtext
                end
              when 'NamedCell'
                if elem4.has_attributes? and (w = elem4.attributes['ss:Name'])
                  hs[key3] = recover_charw(w)  if w != ''
                end
              end
            }
            if sid and styles_rec[sid]
              hs.update(hash_copy(styles_rec[sid]))
            end
            ss.array[row_count-1][cell_count-1] = hs
          }
        end
      }
      ss.source_type = :XMLSS
      ss.source_clue = clue
      sheet_ary << ss
    }
    change_kcode()
    if $ExlapBaseC != 'w'
      for i in 0...sheet_ary.size
        sheet_ary[i].all_code_adj()
      end
    end
    @sheets += sheet_ary
    return self
  end
  alias :push_xml :xml_push
end


# webtableʬФꤹ뤿Υ饹
class ExlWeb < ExlBookBase
  include ExlPrivate

  def initialize(*url_info)
    @sheets = Array.new
    @author = @last_author = @company = nil
    @created = @last_saved = nil
    @styles = ExlStyle.new
    @other = Hash.new
    @other_dp = Hash.new
    if !defined?(Nokogiri) and !defined?(Hpricot)
      begin
        require 'rubygems'
      rescue
        LoadError
      end
      begin
        require 'nokogiri'
      rescue
        LoadError
      end
    end
    if !defined?(Nokogiri) and !defined?(Hpricot) and RUBY_VERSION < "1.9"
      begin
        require 'rubygems'
      rescue
        LoadError
      end
      begin
        require 'hpricot'
      rescue
        LoadError
      end
    end
    if url_info.size > 0
      push(*url_info)
    end
  end

  # ʸ(&#ġ;)ʸѴ
      # ʸɤ $ExlapBaseC (j, e, s, w ʤ)˽
  def ncr_convert(str)
    unless str =~ /\&\#.+;/
      return str
    end
    str = str.gsub(/\&\#x([0-9a-f]+);/i) {|ch|
      n = $1.hex
      ch = "&##{n};"
    }
    str.gsub!(/\&\#(\d+);/) {|ptn|
      n = $1.to_i
      if n > 0xff
        ch = NKF.nkf("-m0W16#{$ExlapBaseC}", [n].pack("n"))
      else
        ch = ptn
      end
      ptn = ch
    }
    return str
  end

  # htmltableʬrubyѴб
      # data_type(γǤΥ)  :TEXT | :HTML
      # :HTMLξ <td>ġ</td> γǤ
      # $cell_count=ֶʳΥ, $merge_count=֥θĿ
  def table_to_array(tbl, data_type=:TEXT)
    aa = []
    $cell_count = $merge_count = 0
    if !defined?(Hpricot) and !defined?(Nokogiri)
      STDERR.puts "can't call \"ExlWeb::table_to_array\" without " +
          "hpricot or nokogiri"
      return aa
    end
    if (e = tbl.at("caption"))
      w = e.inner_text.sub(/^\s+/, "").sub(/\s+$/, "")
      w = "caption:" + w
      aa.push([w])
    end
    trs = tbl.search("tr")
    span = {}
    trs.each_with_index {|e, row_n|
      a = []
      col_n = 0
      e.search("*").each {|e2|
        next  unless e2.name =~ /^(th|td)$/i
        while span[[row_n, col_n]]
          a.push(nil)  # ηսnilˤ
          col_n += 1
        end
        if data_type == :TEXT
          data = e2.inner_text
        else
          data = e2.to_html
        end
        cspan = 1
        if e2['colspan'] =~ /(\d+)/
          cspan = $1.to_i
        end
        rspan = 1
        if e2['rowspan'] =~ /(\d+)/
          rspan = $1.to_i
        end
        if cspan == 1 and rspan == 1  # Ȥ̵ط
          a.push(data)
          $cell_count += 1
        else
          hs = {'Data'=>data, 'Rspan'=>rspan, 'Cspan'=>cspan}
          a.push(hs)
          $cell_count += 1
          $merge_count += 1
          for r in row_n...(row_n+rspan)
            for c in col_n...(col_n+cspan)
              span[[r, c]] = true
            end
          end
        end
        col_n += 1
      }
      aa.push(a)
    }
    return aa
  end

  # ɽʸɤѹ
  def arrange_regex(re, code)
    if re
      str = re.source
      unless str =~ /^[\x00-\x7f]+$/
        opt = re.options
        str = xnkf(str, code)
        re = Regexp.new(str, opt)
      end
    end
    return re
  end

  # ꤵ줿urlwebڡhtmlѡѤ˽
    # url '<' ǻϤޤ html source ʸȤߤʤ
  def get_html(url)
    temp_chr = '__jfZR6XRr__'
    if is_html(url) 
      body = url.sub(/^\s+/, "")
    else
      body = $ExlapMshttp.web_open(url)
    end
    if body == '' or body == false
      STDERR.printf("could not access: %s\n", url)
      return ''
    end
    body = xnkf(body.gsub(/\r\n/, "\n"))
    body = ncr_convert(body)
    body.gsub!(/<pre[^>]*>.+?<\/pre>/im) {|mp|
      mp = mp.gsub(/\n/, "#{temp_chr}")
    }
    body.gsub!(/\n+/, "")
    re = /<(p|br|li|[oud]l|dt|dd|h[1-6]|div|center|blockquote|pre)([ >])/i
    body.gsub!(re, "\n<\\1\\2")
    body.gsub!(/<\/(p|li|[oud]l|dt|dd|h[1-6]|div|center|blockquote|pre)>/i,
        "</\\1>\n")
    body.gsub!(/<(hr|hr +[^>]+)>/i, "\n<\\1>\n")
    body.gsub!(/\n\n+/, "\n")
    body.gsub!(/#{temp_chr}/, "\n")
    body.gsub!(/<img[^>]+>/i) {|mp|  # altִ
      if mp =~ /alt\s*=\s*["'](.+?)["']/im
        w = $1
        w.sub!(/^\s+/, "")
        w.sub!(/\s+$/, "")
        mp = w
      end
    }
    body.gsub!(/\&nbsp;/i, "&#32;")
    return body
  end

  # webtableʬ(˹礦)Ф֤
  def web_to_array(url, table_select_regexp=nil,
      table_select_count=nil)
    res_ary = Array.new
    if !defined?(Hpricot) and !defined?(Nokogiri)
      STDERR.puts "can't call \"ExlWeb::web_to_array\" " +
          "without hpricot or nokogiri"
      return res_ary
    end
    body = get_html(url)
    if body == ''
      return res_ary
    end
    table_select_regexp = arrange_regex(table_select_regexp, $ExlapBaseC)
    change_flag = nil
    if defined?(Hpricot)
      doc = Hpricot(body)
    elsif defined?(Nokogiri)
      change_kcode('u')
      unless $ExlapBaseC == 'w'
        body = xnkf(body, 'w')
        table_select_regexp = arrange_regex(table_select_regexp, 'w')
        change_flag = true
      end
      doc = Nokogiri::HTML.parse(body, nil, "UTF-8")
    end
    doc.search("table").each {|tbl_obj|
      next  if tbl_obj.at("table")  # table¢tableоݳ
      tbl_str = tbl_obj.inner_html  # tableʬhtml source
      aa = table_to_array(tbl_obj)
      next  if aa.size < 1
      if table_select_regexp
        unless table_select_regexp =~ tbl_str
          next
        end
      end
      if table_select_count
        if aa.size < table_select_count
          next
        end
      end
      res_ary << aa
    }
    change_kcode()
    if change_flag
      res_ary.map! {|aa|
        aa = arrange_array_s(aa)
      }
    end
    return res_ary
  end

  # ʣwebtableѴ
  def push(*url_info)
    aa_ary = Array.new
    if defined?(Hpricot) or defined?(Nokogiri)
      while url = url_info.shift
        unless url.class == String
          url_info.shift
          next
        end
        table_select_regexp = table_select_count = nil
        while url_info.size > 0 and url_info[0].class != String
          if url_info[0].class == Regexp
            table_select_regexp = url_info.shift
          elsif int_t(url_info[0])
            table_select_count = url_info.shift
          else
            url_info.shift
          end
        end
        tbls = web_to_array(url, table_select_regexp,
            table_select_count)
        tbls.each {|aa|
          esh = ExlSheet.new(aa)
          esh.source_type = :WEB
          esh.source_clue = url
          aa_ary << esh
        }
      end
      if aa_ary.size > 0
        set_sheet(*aa_ary)
      end
    else  # if !defined?(Hpricot) and !defined?(Nokogiri)
      STDERR.puts "can't call \"ExlWeb::push\" " +
          "without hpricot or nokogiri"
    end
    return self
  end
end


# csv򰷤Υ饹
class ExlCsv < ExlBookBase
  include ExlPrivate

  def initialize(*opt)
    @sheets = Array.new
    @author = @last_author = @company = nil
    @created = @last_saved = nil
    @styles = ExlStyle.new
    @other = Hash.new
    @other_dp = Hash.new
    if opt.size > 0
      push(*opt)
    end
  end

  # csvեޤcsvʸ󤫤
  def push(*opt)
    res = Array.new
    while obj = opt.shift
      if obj.class != String
        opt.shift
        next
      end
      mode = obj.index("\n")  ?  :STRING  :  :FILE
      while opt.size > 0 and opt[0].class == Symbol
        mode = opt.shift
      end
      if mode == :FILE
        filename = obj
        str = File.open(filename) {|ff|  ff.read}
      elsif mode == :STRING
        str = obj
      end
      str = xnkf(str)
      if RUBY_VERSION < "1.9" and defined?(FasterCSV)
        ary = FasterCSV.parse(str)
      else
        ary = CSV.parse(str)
      end
      next  if ary.size < 1
      for y in 0...ary.size
        next  unless ary[y]
        for x in 0...ary[y].size
          ary[y][x] = ary[y][x].to_s  if ary[y][x]
        end
      end
      esh = ExlSheet.new(ary)
      esh.source_type = :CSV
      esh.source_clue = obj
      res << esh
    end
    set_sheet(*res)  if res.size > 0
    return self
  end
end
